import logging
import pytest

from types import SimpleNamespace

from ophyd import EpicsMCA, EpicsDXP
from ophyd.mca import add_rois, Mercury1, SoftDXPTrigger

from ophyd.utils import enum, ReadOnlyError

from caproto.tests.conftest import run_example_ioc

MCAMode = enum(PHA='PHA', MCS='MCS', List='List')
DxpPresetMode = enum(No_preset='No preset',
                     Real_time='Real time',
                     Live_time='Live time')

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def mca_test_ioc(prefix, request):
    name = 'test_signal IOC'
    pvs = dict(mca_prefix=f'{prefix}mca',
               dxp_prefix=f'{prefix}dxp:'
               )

    process = run_example_ioc('ophyd.tests.mca_ioc',
                              request=request,
                              pv_to_check=pvs['mca_prefix'],
                              args=('--prefix', prefix, '-v'))
    return SimpleNamespace(process=process, prefix=prefix, name=name, pvs=pvs,
                           type='caproto')


@pytest.fixture(scope='function')
def mca(cleanup, mca_test_ioc):
    mca = EpicsMCA(mca_test_ioc.pvs['mca_prefix'], name='mca')
    mca.wait_for_connection()
    cleanup.add(mca)
    return mca


@pytest.fixture(scope='function')
def dxp(cleanup, mca_test_ioc):
    dxp = EpicsDXP(mca_test_ioc.pvs['dxp_prefix'], name='dxp')
    dxp.wait_for_connection()
    cleanup.add(dxp)
    return dxp


def test_mca_spectrum(mca):
    with pytest.raises(ReadOnlyError):
        mca.spectrum.put(3.14)
    with pytest.raises(ReadOnlyError):
        mca.background.put(3.14)


def test_mca_read_attrs(mca):
    # default read_attrs
    default_normal_kind = ['preset_real_time', 'elapsed_real_time', 'spectrum']
    assert set(default_normal_kind) == set(mca.read_attrs)
    # test passing in custom read_attrs (with dots!)
    r_attrs = ['spectrum', 'rois.roi1.count', 'rois.roi2.count']

    mca.read_attrs = r_attrs
    expected = set(r_attrs + ['rois.roi1', 'rois.roi2', 'rois'])
    assert expected == set(mca.read_attrs)


def test_mca_describe(mca):
    desc = mca.describe()
    d = desc[mca.name + '_spectrum']

    assert d['dtype'] == 'number'
    assert d['shape'] == []


def test_mca_signals(mca):
    mca.mode.put(MCAMode.PHA)
    mca.stage()
    mca.start.put(1)
    mca.stop_signal.put(1)
    mca.preset_real_time.put(3.14)
    mca.preset_live_time.put(3.14)
    mca.erase_start.put(1)
    mca.stop()
    mca.unstage()


def test_rois(mca):
    # iterables only
    with pytest.raises(TypeError):
        add_rois(1)
    # check range
    with pytest.raises(ValueError):
        add_rois([-1, ])
    with pytest.raises(ValueError):
        add_rois([32, ])
    # read-only?
    with pytest.raises(ReadOnlyError):
        mca.rois.roi1.count.put(3.14)
    with pytest.raises(ReadOnlyError):
        mca.rois.roi1.net_count.put(3.14)


def test_dxp_signals(dxp):
    # NOTE: values used below are those currently used at 23id2
    dxp.preset_mode.put(DxpPresetMode.Real_time)
    dxp.stage()
    dxp.unstage()

    dxp.trigger_peaking_time.put(0.2)
    dxp.trigger_threshold.put(0.6)
    dxp.trigger_gap_time.put(0.0)
    dxp.max_width.put(1.0)
    dxp.peaking_time.put(0.25)
    dxp.energy_threshold.put(0.35)
    dxp.gap_time.put(0.05)

    dxp.baseline_cut_percent.put(5.0)
    dxp.baseline_cut_enable.put(1)
    dxp.baseline_filter_length.put(128)
    dxp.baseline_threshold.put(0.20)

    dxp.preamp_gain.put(5.5)
    dxp.detector_polarity.put(1)
    dxp.reset_delay.put(10.0)
    dxp.decay_time.put(50.0)
    dxp.max_energy.put(2.0)
    dxp.adc_percent_rule.put(5.0)

    # read-only
    with pytest.raises(ReadOnlyError):
        dxp.triggers.put(2)
    with pytest.raises(ReadOnlyError):
        dxp.events.put(2)
    with pytest.raises(ReadOnlyError):
        dxp.overflows.put(2)
    with pytest.raises(ReadOnlyError):
        dxp.underflows.put(2)
    with pytest.raises(ReadOnlyError):
        dxp.input_count_rate.put(2)
    with pytest.raises(ReadOnlyError):
        dxp.output_count_rate.put(2)


def test_mixin_signals():
    class Mercury(Mercury1, SoftDXPTrigger):
        pass
    mercury = Mercury('Will_Not_Try_To_connect', name='mercury')
    assert 'count_time' in mercury.component_names



import logging
import pytest
import time

from copy import copy
from types import SimpleNamespace

from ophyd.scaler import EpicsScaler, ScalerCH
from ophyd.utils import enum
from .test_utils import assert_OD_equal_ignore_ts

from caproto.tests.conftest import run_example_ioc

ScalerMode = enum(ONE_SHOT=0, AUTO_COUNT=1)

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def scaler_test_ioc(prefix, request):
    name = 'test_signal IOC'
    pvs = dict(scaler_prefix=f'{prefix}')

    process = run_example_ioc('ophyd.tests.scaler_ioc',
                              request=request,
                              pv_to_check=pvs['scaler_prefix'] + '.CNT',
                              args=('--prefix', prefix, '-v'))
    return SimpleNamespace(process=process, prefix=prefix, name=name, pvs=pvs,
                           type='caproto')


@pytest.fixture(scope='function')
def scaler(cleanup, scaler_test_ioc):
    scaler = EpicsScaler(scaler_test_ioc.pvs['scaler_prefix'], name='scaler')
    scaler.wait_for_connection()
    cleanup.add(scaler)
    return scaler


@pytest.fixture(scope='function')
def scaler_ch(cleanup, scaler_test_ioc):
    ch = ScalerCH(scaler_test_ioc.pvs['scaler_prefix'])
    ch.wait_for_connection()
    cleanup.add(scaler)
    return ch


def test_scaler_connects(scaler):
    ...


def test_scaler_functionality(scaler):
    scaler.read_attrs = ['channels']

    scaler.wait_for_connection()
    scaler.count_mode.put('OneShot')
    scaler.preset_time.put(5.2)

    logger.info('Counting in One-Shot mode for %f s...',
                scaler.preset_time.get())
    scaler.count.put(1)
    logger.info('Sleeping...')
    time.sleep(0.1)
    logger.info('Done sleeping. Stopping counter...')
    scaler.stop()

    logger.info('Set mode to AutoCount')
    scaler.count_mode.put('AutoCount')
    scaler.count.put(1)
    logger.info('Begin auto-counting (aka "background counting")...')
    time.sleep(0.1)
    logger.info('Set mode to OneShot')
    scaler.count_mode.put('OneShot')
    time.sleep(0.1)
    logger.info('Stopping (aborting) auto-counting.')
    scaler.stop()

    logger.info('read() all channels in one-shot mode...')
    vals = scaler.read()
    logger.info(vals)
    assert 'scaler_channels_chan1' in vals

    scaler.report
    scaler.read()

    assert copy(scaler).prefix == scaler.prefix
    assert copy(scaler).read_attrs == scaler.read_attrs
    assert copy(scaler).configuration_attrs == scaler.configuration_attrs
    repr(scaler)
    str(scaler)

    scaler.stage()
    old, new = scaler.configure({})
    scaler.unstage()

    assert_OD_equal_ignore_ts(old, new)

    scaler.stage()
    old_preset_time = scaler.preset_time.get()
    old, new = scaler.configure({'preset_time': 7})
    scaler.unstage()

    assert old.pop('scaler_preset_time')['value'] == old_preset_time
    assert new.pop('scaler_preset_time')['value'] == 7
    assert_OD_equal_ignore_ts(old, new)

    scaler.hints == {'fields': [scaler.channels.name]}


def test_signal_separate(scaler):
    scaler.read_attrs = ['channels.chan1']

    data = scaler.read()
    assert 'scaler_channels_chan1' in data
    assert 'scaler_channels_chan2' not in data


def smoke_test_scalerCH(scaler_ch):
    data = scaler_ch.read()
    assert 'scaler_channels_chan1' in data
    assert 'scaler_channels_chan2' not in data



from .. import Device, Component as Cpt
from . import plugins
from ..device import create_device_from_components
from ..ophydobj import select_version, OphydObject


class CommonPlugins(Device, version_type='ADCore'):
    ...


class CommonOverlayPlugin(Device, version_type='ADCore'):
    ...


class CommonAttributePlugin(Device, version_type='ADCore'):
    ...


class CommonROIStatPlugin(Device, version_type='ADCore'):
    ...


class CommonGatherPlugin(Device, version_type='ADCore'):
    ...


def _get_bases(cls, version):
    mixin_cls = select_version(cls, version)
    base_cls = select_version(plugins.PluginBase, version)
    if issubclass(mixin_cls, base_cls):
        return (mixin_cls, )

    return (mixin_cls, base_cls)


def _make_common_numbered(clsname, version, cpt_cls_base, bases,
                          attr_prefix):
    try:
        cpt_cls = select_version(cpt_cls_base, version)
    except ValueError:
        return

    base_cls_base, plugin_base = bases
    bases = (base_cls_base, ) + _get_bases(plugin_base, version)

    return create_device_from_components(
        name=clsname,
        base_class=bases,
        class_kwargs={
            'version_of': base_cls_base,
            'version': version},
        **{f'{attr_prefix}_{j}': Cpt(cpt_cls, f'{j}:')
           for j in range(1, 9)}
    )


def _make_common_gather(clsname, version):
    base_cls_base = plugins.GatherPlugin
    cpt_cls_base = plugins.GatherNPlugin

    try:
        cpt_cls = select_version(cpt_cls_base, version)
    except ValueError:
        return

    bases = (CommonGatherPlugin, ) + _get_bases(base_cls_base, version)

    return create_device_from_components(
        name=clsname,
        base_class=bases,
        class_kwargs={
            'version_of': CommonGatherPlugin,
            'version': version},
        **{f'gather_{j}': Cpt(cpt_cls, '', index=j)
           for j in range(1, 9)}
    )


versions = [(1, 9, 1),
            (2, 0),
            (2, 1),
            (2, 2),
            (2, 3),
            (2, 4),
            (2, 5),
            (2, 6),
            (3, 1),
            (3, 2),
            (3, 3),
            (3, 4)
            ]


for version in versions:
    ver_string = "".join(str(_) for _ in version)
    _make_common_numbered(f'CommonOverlayPlugin_V{ver_string}',
                          version,
                          plugins.Overlay, (CommonOverlayPlugin,
                                            plugins.OverlayPlugin,),
                          'overlay')
    _make_common_numbered(f'CommonAttributePlugin_V{ver_string}',
                          version,
                          plugins.AttributeNPlugin, (CommonAttributePlugin,
                                                     plugins.AttributePlugin),
                          'attr')
    _make_common_numbered(f'CommonROIStatPlugin_V{ver_string}',
                          version,
                          plugins.ROIStatNPlugin, (CommonROIStatPlugin,
                                                   plugins.ROIStatPlugin),
                          'attr')

    _make_common_gather(f'CommonGatherPlugin_V{ver_string}', version)


all_plugins = [
    ('attr1', CommonAttributePlugin, 'Attr1:'),
    ('cb1', plugins.CircularBuffPlugin, 'CB1:'),
    ('cc1', plugins.ColorConvPlugin, 'CC1:'),
    ('cc2', plugins.ColorConvPlugin, 'CC2:'),
    ('codec1', plugins.CodecPlugin, 'Codec1:'),
    ('codec2', plugins.CodecPlugin, 'Codec2:'),
    ('fft1', plugins.FFTPlugin, 'FFT1:'),
    ('gather1', CommonGatherPlugin, 'Gather1:'),
    ('hdf1', plugins.HDF5Plugin, 'HDF1:'),
    ('jpeg1', plugins.JPEGPlugin, 'JPEG1:'),
    ('magick1', plugins.MagickPlugin, 'Magick1:'),
    ('netcdf1', plugins.NetCDFPlugin, 'netCDF1:'),
    ('nexus1', plugins.NexusPlugin, 'Nexus1:'),
    ('over1', CommonOverlayPlugin, 'Over1:'),
    ('proc1', plugins.ProcessPlugin, 'Proc1:'),
    ('proc1_tiff', plugins.TIFFPlugin, 'Proc1:TIFF:'),
    ('roi1', plugins.ROIPlugin, 'ROI1:'),
    ('roi2', plugins.ROIPlugin, 'ROI2:'),
    ('roi3', plugins.ROIPlugin, 'ROI3:'),
    ('roi4', plugins.ROIPlugin, 'ROI4:'),
    ('roistat1', CommonROIStatPlugin, 'ROIStat1:'),
    ('scatter1', plugins.ScatterPlugin, 'Scatter1:'),
    ('stats1', plugins.StatsPlugin, 'Stats1:'),
    ('stats1_ts', plugins.TimeSeriesPlugin, 'Stats1:TS:'),
    ('stats2', plugins.StatsPlugin, 'Stats2:'),
    ('stats2_ts', plugins.TimeSeriesPlugin, 'Stats2:TS:'),
    ('stats3', plugins.StatsPlugin, 'Stats3:'),
    ('stats3_ts', plugins.TimeSeriesPlugin, 'Stats3:TS:'),
    ('stats4', plugins.StatsPlugin, 'Stats4:'),
    ('stats4_ts', plugins.TimeSeriesPlugin, 'Stats4:TS:'),
    ('stats5', plugins.StatsPlugin, 'Stats5:'),
    ('stats5_ts', plugins.TimeSeriesPlugin, 'Stats5:TS:'),
    ('tiff1', plugins.TIFFPlugin, 'TIFF1:'),
    ('trans1', plugins.TransformPlugin, 'Trans1:')
]


common_plugins = {}


for version in versions:
    local_plugins = {}
    for attr, cls, suffix in all_plugins:
        if attr == 'magick1' and version > (3, 0):
            continue
        elif attr.endswith('_ts') and version < (3, 3):
            continue
        elif attr == 'fft1' and version < (3, 0):
            continue
        elif attr == 'proc1_tiff' and version < (3, 3):
            continue
        elif attr == 'attr1' and version < (2, 2):
            continue

        try:
            cls = select_version(cls, version)
        except ValueError:
            continue

        local_plugins[attr] = Cpt(cls, suffix)

    ver_string = "".join(str(_) for _ in version)
    class_name = f'CommonPlugins_V{ver_string}'

    common_plugins[class_name] = create_device_from_components(
        name=class_name,
        base_class=CommonPlugins,
        class_kwargs=dict(
            version=version,
            version_of=CommonPlugins),
        **local_plugins)


versioned_plugins = {}


class PluginNamespace(OphydObject, version_type='ADCore'):
    """
    This is intended to be used with select_versions.

    This provides namespaces of plugins for a given version of Area Detector.

    Examples
    --------
    Access the StatsPlugin for Area Detector version 3.4.

    >>> select_version(PluginNamespace, (3, 4)).StatsPlugin
    ophyd.areadetector.plugins.StatsPlugin_V34
    """
    # This is implemented as an OphydObject in order to leverage OphydObject's
    # version-resolution class keyword arguments. But it should *not* be used
    # as a Device or Signal. It should be used with select_version as in the
    # example in the docstring.
    ...


for version in versions:
    local_plugins = {}
    for _, cls, _ in all_plugins:
        try:
            local_plugins[cls.__name__] = select_version(cls, version)
        except ValueError:
            continue
    ver_string = "".join(str(_) for _ in version)
    class_name = f"PluginNamespace_V{ver_string}"
    versioned_plugins[class_name] = type(class_name,
                                         (PluginNamespace,),
                                         local_plugins,
                                         version=version,
                                         version_of=PluginNamespace)


globals().update(**common_plugins, **versioned_plugins)
del local_plugins
del version
del class_name
del suffix
del ver_string
del attr
del cls


__all__ = ['PluginNamespace', 'CommonPlugins']



import time
import pytest

from ophyd import (Component as Cpt,
                   SimDetector, SimDetectorCam, StatsPlugin, EpicsMotor,
                   ROIPlugin, ProcessPlugin, Device)
from ophyd.areadetector.base import EpicsSignalWithRBV
from ophyd.flyers import (AreaDetectorTimeseriesCollector,
                          WaveformCollector,
                          MonitorFlyerMixin)
from ophyd.status import wait


@pytest.fixture(scope='function')
def ts_sim_detector(ad_prefix):
    class Detector(SimDetector):
        acquire = Cpt(EpicsSignalWithRBV, 'cam1:Acquire', trigger_value=1)
        cam = Cpt(SimDetectorCam, 'cam1:')
        ts_col = Cpt(AreaDetectorTimeseriesCollector, 'Stats1:')
        stats = Cpt(StatsPlugin, 'Stats1:')
        roi1 = Cpt(ROIPlugin, 'ROI1:')
        proc1 = Cpt(ProcessPlugin, 'Proc1:')

    try:
        det = Detector(ad_prefix, name='sim')
        det.wait_for_connection(timeout=1.0)
    except TimeoutError:
        pytest.skip('IOC unavailable')
    return det


@pytest.fixture
def tscollector(ts_sim_detector):
    return ts_sim_detector.ts_col


@pytest.mark.adsim
def test_ad_time_series(ts_sim_detector, tscollector):
    sim_detector = ts_sim_detector

    num_points = 3

    cam = sim_detector.cam
    cam.stage_sigs[cam.acquire_time] = 0.001
    cam.stage_sigs[cam.acquire_period] = 0.001
    cam.stage_sigs[cam.image_mode] = 'Single'
    cam.stage_sigs[cam.trigger_mode] = 'Internal'

    print('tscollector desc', tscollector.describe())
    print('tscollector flyer desc', tscollector.describe_collect())
    print('tscollector repr', repr(tscollector))
    print('simdet stage sigs', sim_detector.stage_sigs)
    print('tscoll stage sigs', tscollector.stage_sigs)
    print('cam stage sigs', cam.stage_sigs)
    print('stats stage sigs', sim_detector.stats.stage_sigs)

    try:
        # In case the collector is currently running...
        wait(tscollector.complete())
    except RuntimeError:
        ...

    tscollector.stage_sigs[tscollector.num_points] = num_points

    sim_detector.stage()
    tscollector.kickoff()

    for i in range(num_points):
        st = sim_detector.trigger()
        wait(st)
        print(st)
        time.sleep(0.1)

    tscollector.pause()
    tscollector.resume()

    st = tscollector.complete()
    wait(st)

    collected = list(tscollector.collect())
    print('collected', collected)
    sim_detector.unstage()
    assert len(collected) == num_points
    # TODO any more validation here?


@pytest.fixture(scope='function')
def wf_sim_detector(ad_prefix):
    suffix = '??TODO??'

    class Detector(SimDetector):
        wfcol = Cpt(WaveformCollector, suffix)

    try:
        det = Detector(ad_prefix, name='det')
        det.wait_for_connection(timeout=1.0)
    except TimeoutError:
        pytest.skip('IOC unavailable')
    return det


@pytest.fixture
def wfcol(wf_sim_detector):
    return wf_sim_detector.wfcol


@pytest.mark.adsim
def test_waveform(wf_sim_detector, wfcol):
    print('waveform collector', wfcol)
    print('wfcol flyer desc', wfcol.describe_collect())


@pytest.mark.motorsim
@pytest.mark.parametrize('pivot', [True, False])
def test_monitor_flyer(pivot):
    class BasicDevice(Device):
        mtr1 = Cpt(EpicsMotor, 'XF:31IDA-OP{Tbl-Ax:X2}Mtr')
        mtr2 = Cpt(EpicsMotor, 'XF:31IDA-OP{Tbl-Ax:X3}Mtr')

    class FlyerDevice(MonitorFlyerMixin, BasicDevice):
        pass

    fdev = FlyerDevice('', name='fdev', stream_names={'mtr1.user_readback': 'oranges'},
                       pivot=pivot)
    fdev.wait_for_connection()

    fdev.monitor_attrs = ['mtr1.user_readback', 'mtr2.user_readback']
    fdev.describe()

    st = fdev.kickoff()
    wait(st)

    mtr1, mtr2 = fdev.mtr1, fdev.mtr2
    rbv1, rbv2 = mtr1.position, mtr2.position
    fdev.mtr1.move(rbv1 + 0.2, wait=True)
    fdev.mtr2.move(rbv2 + 0.2, wait=True)

    fdev.pause()

    fdev.mtr1.move(rbv1 - 0.2, wait=True)
    fdev.mtr2.move(rbv2 - 0.2, wait=True)

    fdev.resume()
    st = fdev.complete()
    wait(st)

    print(fdev.describe_collect())

    desc1 = fdev.mtr1.user_readback.describe()
    desc2 = fdev.mtr2.user_readback.describe()

    if not pivot:
        desc1[fdev.mtr1.user_readback.name]['dtype'] = 'array'
        desc2[fdev.mtr2.user_readback.name]['dtype'] = 'array'

    assert (fdev.describe_collect() ==
            {'oranges': desc1,
             'fdev_mtr2': desc2,
             }
            )
    data = list(fdev.collect())
    print('collected data', data)
    # data from both motors

    if not pivot:
        assert len(data) == 2
        d1 = data[0]['data']['fdev_mtr1']
        d2 = data[1]['data']['fdev_mtr2']

    else:
        assert len(data) >= 2

        d1 = [d['data']['fdev_mtr1'] for d in data
              if 'fdev_mtr1' in d['data']]
        d2 = [d['data']['fdev_mtr2'] for d in data
              if 'fdev_mtr2' in d['data']]

    # and at least more than one data point...
    assert len(d1) > 1
    assert len(d2) > 1

    print('data1', d1)
    print('data2', d2)
    # raise ValueError()



import atexit
import logging
from distutils.version import LooseVersion
import epics
from epics import ca, caget, caput

from ._dispatch import _CallbackThread, EventDispatcher, wrap_callback

_min_pyepics = '3.4.2'

if LooseVersion(epics.__version__) < LooseVersion(_min_pyepics):
    raise ImportError('Version of pyepics too old. '
                      f'Ophyd requires at least {_min_pyepics}'
                      f'but {epics.__version__} is installed')

try:
    ca.find_libca()
except ca.ChannelAccessException:
    raise ImportError('libca not found; pyepics is unavailable')
else:
    thread_class = ca.CAThread


module_logger = logging.getLogger(__name__)
name = 'pyepics'
_dispatcher = None
get_pv = epics.get_pv


def get_dispatcher():
    'The event dispatcher for the pyepics control layer'
    return _dispatcher


class PyepicsCallbackThread(_CallbackThread):
    def attach_context(self):
        super().attach_context()
        ca.attach_context(self.context)

    def detach_context(self):
        super().detach_context()
        if ca.current_context() is not None:
            ca.detach_context()


class PyepicsShimPV(epics.PV):
    def __init__(self, pvname, callback=None, form='time', verbose=False,
                 auto_monitor=None, count=None, connection_callback=None,
                 connection_timeout=None, access_callback=None):
        connection_callback = wrap_callback(_dispatcher, 'metadata',
                                            connection_callback)
        callback = wrap_callback(_dispatcher, 'monitor', callback)
        access_callback = wrap_callback(_dispatcher, 'metadata',
                                        access_callback)

        super().__init__(pvname, form=form, verbose=verbose,
                         auto_monitor=auto_monitor, count=count,
                         connection_timeout=connection_timeout,
                         connection_callback=connection_callback,
                         callback=callback, access_callback=access_callback)

        self._cache_key = (pvname, form, self.context)
        self._reference_count = 0

    def add_callback(self, callback=None, index=None, run_now=False,
                     with_ctrlvars=True, **kw):
        if not self.auto_monitor:
            self.auto_monitor = ca.DEFAULT_SUBSCRIPTION_MASK
        callback = wrap_callback(_dispatcher, 'monitor', callback)
        return super().add_callback(callback=callback, index=index,
                                    run_now=run_now,
                                    with_ctrlvars=with_ctrlvars, **kw)

    def put(self, value, wait=False, timeout=None, use_complete=False,
            callback=None, callback_data=None):
        callback = wrap_callback(_dispatcher, 'get_put', callback)
        # pyepics does not accept an indefinite timeout
        if timeout is None:
            timeout = 315569520  # ten years
        return super().put(value, wait=wait, timeout=timeout,
                           use_complete=use_complete, callback=callback,
                           callback_data=callback_data)

    def _getarg(self, arg):
        "wrapper for property retrieval"
        # NOTE: replaces epics.PV._getarg: does not call get() when value unset
        if self._args[arg] is None:
            if arg in ('status', 'severity', 'timestamp', 'posixseconds',
                       'nanoseconds'):
                self.get_timevars(timeout=1, warn=False)
            else:
                self.get_ctrlvars(timeout=1, warn=False)
        return self._args.get(arg, None)

    def get_all_metadata_blocking(self, timeout):
        if self._args['status'] is None:
            self.get_timevars(timeout=timeout)
        self.get_ctrlvars(timeout=timeout)
        md = self._args.copy()
        md.pop('value', None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        _dispatcher.schedule_utility_task(get_metadata_thread,
                                          pvname=self.pvname)

    def clear_callbacks(self):
        super().clear_callbacks()
        self.access_callbacks.clear()
        self.connection_callbacks.clear()


def release_pvs(*pvs):
    for pv in pvs:
        pv._reference_count -= 1
        if pv._reference_count == 0:
            pv.clear_callbacks()
            pv.clear_auto_monitor()
            if pv.chid is not None:
                # Clear the channel on the CA-level
                epics.ca.clear_channel(pv.chid)

            pv.chid = None
            pv.context = None

            # Ensure we don't get this same PV back again
            epics.pv._PVcache_.pop(pv._cache_key, None)


def setup(logger):
    '''Setup ophyd for use

    Must be called once per session using ophyd
    '''
    # It's important to use the same context in the callback _dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug('ophyd already setup')
        return

    epics.pv.default_pv_class = PyepicsShimPV

    def _cleanup():
        '''Clean up the ophyd session'''
        global _dispatcher
        if _dispatcher is None:
            return
        epics.pv.default_pv_class = epics.PV

        if _dispatcher.is_alive():
            _dispatcher.stop()

        _dispatcher = None

    logger.debug('Installing event dispatcher')
    _dispatcher = EventDispatcher(thread_class=PyepicsCallbackThread,
                                  context=ca.current_context(), logger=logger)
    atexit.register(_cleanup)
    return _dispatcher


__all__ = ('setup', 'caput', 'caget', 'get_pv', 'thread_class', 'name',
           'release_pvs', 'get_dispatcher')



import threading

import atexit
import logging

from caproto.threading import pyepics_compat
from caproto.threading.pyepics_compat import PV as _PV, caput, caget  # noqa
from ._dispatch import _CallbackThread, EventDispatcher, wrap_callback


thread_class = threading.Thread
module_logger = logging.getLogger(__name__)
_dispatcher = None
name = 'caproto'


def get_dispatcher():
    'The event dispatcher for the caproto control layer'
    return _dispatcher


class CaprotoCallbackThread(_CallbackThread):
    ...


class PV(_PV):
    def __init__(self, pvname, callback=None, form='time', verbose=False,
                 auto_monitor=None, count=None, connection_callback=None,
                 connection_timeout=None, access_callback=None,
                 context=None):
        connection_callback = wrap_callback(_dispatcher, 'metadata',
                                            connection_callback)
        callback = wrap_callback(_dispatcher, 'monitor', callback)
        access_callback = wrap_callback(_dispatcher, 'metadata',
                                        access_callback)

        super().__init__(pvname, form=form, verbose=verbose,
                         auto_monitor=auto_monitor, count=count,
                         connection_timeout=connection_timeout,
                         connection_callback=connection_callback,
                         callback=callback, access_callback=access_callback,
                         context=context)

    def add_callback(self, callback=None, index=None, run_now=False,
                     with_ctrlvars=True, **kw):
        if not self.auto_monitor:
            self.auto_monitor = True
        callback = wrap_callback(_dispatcher, 'monitor', callback)
        return super().add_callback(callback=callback, index=index,
                                    run_now=run_now,
                                    with_ctrlvars=with_ctrlvars, **kw)

    def put(self, value, wait=False, timeout=30.0, use_complete=False,
            callback=None, callback_data=None):
        if callback:
            use_complete = True
        callback = wrap_callback(_dispatcher, 'get_put', callback)
        return super().put(value, wait=wait, timeout=timeout,
                           use_complete=use_complete, callback=callback,
                           callback_data=callback_data)

    # TODO: caproto breaks API compatibility in wait_for_connection, raising TimeoutError

    def get_all_metadata_blocking(self, timeout):
        if self._args['status'] is None:
            self.get_timevars(timeout=timeout)
        self.get_ctrlvars(timeout=timeout)
        md = self._args.copy()
        md.pop('value', None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        _dispatcher.schedule_utility_task(get_metadata_thread,
                                          pvname=self.pvname)

    def clear_callbacks(self):
        super().clear_callbacks()
        self.access_callbacks.clear()
        self.connection_callbacks.clear()

    def clear_auto_monitor(self):
        # TODO move into caproto
        self.auto_monitor = False
        if self._auto_monitor_sub is not None:
            self._auto_monitor_sub.clear()
            self._auto_monitor_sub = None


def release_pvs(*pvs):
    for pv in pvs:
        pv.clear_callbacks()
        # pv.disconnect()


def get_pv(pvname, form='time', connect=False, context=None, timeout=5.0,
           connection_callback=None, access_callback=None, callback=None,
           **kwargs):
    """Get a PV from PV cache or create one if needed.

    Parameters
    ---------
    form : str, optional
        PV form: one of 'native' (default), 'time', 'ctrl'
    connect : bool, optional
        whether to wait for connection (default False)
    context : int, optional
        PV threading context (defaults to current context)
    timeout : float, optional
        connection timeout, in seconds (default 5.0)
    """
    if context is None:
        context = PV._default_context

    pv = PV(pvname, form=form, connection_callback=connection_callback,
            access_callback=access_callback, callback=callback,
            **kwargs)
    pv._reference_count = 0
    if connect:
        pv.wait_for_connection(timeout=timeout)
    return pv


def setup(logger):
    '''Setup ophyd for use

    Must be called once per session using ophyd
    '''
    # It's important to use the same context in the callback _dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug('ophyd already setup')
        return

    pyepics_compat._get_pv = pyepics_compat.get_pv
    pyepics_compat.get_pv = get_pv

    def _cleanup():
        '''Clean up the ophyd session'''
        global _dispatcher
        if _dispatcher is None:
            return

        pyepics_compat.get_pv = pyepics_compat._get_pv

        if _dispatcher.is_alive():
            _dispatcher.stop()

        _dispatcher = None

    logger.debug('Installing event dispatcher')
    context = PV._default_context.broadcaster
    _dispatcher = EventDispatcher(thread_class=CaprotoCallbackThread,
                                  context=context, logger=logger)
    atexit.register(_cleanup)
    return _dispatcher
